#include "Synthesizer.hpp"

#include "Keywords.hpp"
#include "Instruction.hpp"
#include "Error.hpp"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

namespace SoftWire
{
	Synthesizer::Synthesizer()
	{
		reset();
	}

	Synthesizer::~Synthesizer()
	{
	}

	void Synthesizer::reset()
	{
		encoding.reset();

		firstType = Operand::UNKNOWN;
		secondType = Operand::UNKNOWN;

		firstReg = Encoding::REG_UNKNOWN;
		secondReg = Encoding::REG_UNKNOWN;
		baseReg = Encoding::REG_UNKNOWN;
		indexReg = Encoding::REG_UNKNOWN;
		
		scale = 0;
	}

	void Synthesizer::defineLabel(const char *label)
	{
		if(encoding.label != 0)
		{
			throw Error::INTERNAL;   // Parser error
		}

		encoding.setLabel(label);
	}

	void Synthesizer::referenceLabel(const char *label)
	{
		if(encoding.reference != 0)
		{
			throw Error("Instruction can't have multiple references");
		}

		encoding.setReference(label);
	}

	void Synthesizer::encodeFirstOperand(const Operand &firstOperand)
	{
		if(firstType != Operand::UNKNOWN)
		{
			throw Error::INTERNAL;   // Instrucition destination already set
		}
	
		firstType = firstOperand.type;

		if(Operand::isReg(firstType) || Operand::isMem(firstType))
		{
			firstReg = firstOperand.reg;
		}
		else if(Operand::isImm(firstType))
		{
			encodeImmediate(firstOperand.value);
		}
		else if(!Operand::isVoid(firstType))
		{
			throw Error::INTERNAL;
		}
	}

	void Synthesizer::encodeSecondOperand(const Operand &secondOperand)
	{
		if(secondType != Operand::UNKNOWN)
		{
			throw Error::INTERNAL;   // Instruction source already set
		}

		secondType = secondOperand.type;

		if(Operand::isReg(secondType) || Operand::isMem(secondType))
		{
			secondReg = secondOperand.reg;
		}
		else if(Operand::isImm(secondType))
		{
			encodeImmediate(secondOperand.value);
		}
		else if(!Operand::isVoid(secondType))
		{
			throw Error::INTERNAL;
		}
	}

	void Synthesizer::encodeThirdOperand(const Operand &thirdOperand)
	{
		if(Operand::isImm(thirdOperand))
		{
			encodeImmediate(thirdOperand.value);
		}
		else if(!Operand::isVoid(thirdOperand))
		{
			throw Error::INTERNAL;
		}
	}

	void Synthesizer::encodeBase(const Operand &base)
	{
		if(baseReg != Encoding::REG_UNKNOWN)
		{
			// Base already set, use as index with scale = 1
			encodeIndex(base);
			encodeScale(1);
			return;
		}

		baseReg = base.reg;
	}

	void Synthesizer::encodeIndex(const Operand &index)
	{
		if(indexReg != Encoding::REG_UNKNOWN)
		{
			throw Error("Memory reference can't have multiple index registers");
		}

		indexReg = index.reg;
	}
	
	void Synthesizer::encodeScale(int scale)
	{
		if(this->scale != 0)
		{
			throw Error("Memory reference can't have multiple scale factors");
		}

		if(scale != 1 && scale != 2 && scale != 4 && scale != 8)
		{
			throw Error("Invalid scale value '%d'", scale);
		}

		this->scale = scale;
	}

	void Synthesizer::encodeImmediate(int immediate)
	{
		if(encoding.immediate != 0)
		{
			throw Error("Instruction can't have multiple immediate operands");
		}

		encoding.immediate = immediate;
	}

	void Synthesizer::encodeDisplacement(int displacement)
	{
		encoding.displacement += displacement;
	}

	const Encoding &Synthesizer::encodeInstruction(const Instruction *instruction)
	{
		if(!instruction)
		{
			return encoding;
		}

		const char *format = instruction->getEncoding();

		if(!format)
		{
			throw Error::INTERNAL;
		}

		while(*format)
		{
			switch((short)format[1] | format[0] << 8)   // Swap byte order so we can use character constants
			{
			case 'p0':
				encoding.addPrefix(0xF0);
				break;
			case 'p2':
				encoding.addPrefix(0xF2);
				break;
			case 'p3':
				encoding.addPrefix(0xF3);
				break;
			case 'po':
				if(!instruction->is32Bit())
				{
					encoding.addPrefix(0x66);
				}
				break;
			case 'pa':
				if(!instruction->is32Bit())
				{
					encoding.addPrefix(0x67);
				}
				break;
			case '+r':
				if(encoding.format.O1)
				{
					if(Operand::isReg(firstType))
					{
						encoding.O1 += firstReg;
					}
					else if(Operand::isReg(secondType))
					{
						encoding.O1 += secondReg;
					}
					else
					{
						throw Error::INTERNAL;   // '+r' not compatible with operands
					}
				}
				else
				{
					throw Error::INTERNAL;   // '+r' needs first opcode byte
				}
				break;
			case '/r':
				encodeModField();

				if(Operand::isReg(instruction->getFirstOperand()) &&
				   Operand::isR_M(instruction->getSecondOperand()))
				{
					if(Operand::isMem(secondType))
					{
						encoding.modRM.r_m = baseReg;
					}
					else if(Operand::isReg(secondType))
					{
						encoding.modRM.r_m = secondReg;
					}
					else
					{
						throw Error::INTERNAL;   // Syntax error should be detected by parser
					}

					encoding.modRM.reg = firstReg;
				}
				else if(Operand::isR_M(instruction->getFirstOperand()) &&
				        Operand::isReg(instruction->getSecondOperand()))
				{
					if(Operand::isMem(firstType))
					{
						encoding.modRM.r_m = baseReg;
					}
					else if(Operand::isReg(firstType))
					{
						encoding.modRM.r_m = firstReg;
					}
					else
					{
						throw Error::INTERNAL;   // Syntax error should be detected by parser
					}

					encoding.modRM.reg = secondReg;
				}
				else
				{
					throw Error::INTERNAL;
				}

				encodeSibByte();

				break;
			case '/0':
			case '/1':
			case '/2':
			case '/3':
			case '/4':
			case '/5':
			case '/6':
			case '/7':
				encodeModField();

				encoding.modRM.reg = (Encoding::Reg)(format[1] - '0');

				if(Operand::isMem(firstType))
				{
					encoding.modRM.r_m = baseReg;
				}
				else if(Operand::isReg(firstType))
				{
					encoding.modRM.r_m = firstReg;
				}
				else
				{
					throw Error::INTERNAL;   // Syntax error should be caught by parser
				}

				encodeSibByte();

				break;
			case 'id':
				encoding.format.I1 = true;
				encoding.format.I2 = true;
				encoding.format.I3 = true;
				encoding.format.I4 = true;
				break;
			case 'iw':
				encoding.format.I1 = true;
				encoding.format.I2 = true;
				break;
			case 'ib':
				encoding.format.I1 = true;
				break;
			case '-b':
				encoding.format.I1 = true;
				break;
			case '-i':
				// Assume 32-bit code
				encoding.format.I1 = true;
				encoding.format.I2 = true;
				encoding.format.I3 = true;
				encoding.format.I4 = true;
				break;
			default:
				unsigned int O = strtoul(format, 0, 16);

				if(O > 0xFF)
				{
					throw Error::INTERNAL;
				}

				if(!encoding.format.O1)
				{
					encoding.O1 = (byte)O;

					encoding.format.O1 = true;
				}
				else if(!encoding.format.O2 &&
				        (encoding.O1 == 0x0F ||
				         encoding.O1 == 0xD8 ||
				         encoding.O1 == 0xD9 ||
				         encoding.O1 == 0xDA ||
				         encoding.O1 == 0xDB ||
				         encoding.O1 == 0xDC ||
				         encoding.O1 == 0xDD ||
				         encoding.O1 == 0xDE ||
				         encoding.O1 == 0xDF))
				{
					encoding.O2 = encoding.O1;
					encoding.O1 = O;

					encoding.format.O2 = true;
				}
				else
				{
					throw Error::INTERNAL;
				}
			}

			format += 2;

			if(*format == ' ')	
			{
				format++;
			}
			else if(*format == '\0')
			{
				break;
			}
			else
			{
				throw Error::INTERNAL;
			}
		}

		return encoding;
	}

	void Synthesizer::encodeModField()
	{
		encoding.format.modRM = true;
		
		if(Operand::isReg(firstType) &&
		   (Operand::isReg(secondType) || Operand::isImm(secondType) || Operand::isVoid(secondType)))
		{
			encoding.modRM.mod = Encoding::MOD_REG;
		}
		else if((Operand::isMem(firstType) || Operand::isMem(secondType)) &&
		        (Operand::isReg(firstType) || Operand::isReg(secondType)))
		{
			if(!encoding.displacement)
			{
				encoding.modRM.mod = Encoding::MOD_NO_DISP;
			}
			else if((char)encoding.displacement == encoding.displacement)
			{
				encoding.modRM.mod = Encoding::MOD_BYTE_DISP;
				encoding.format.D1 = true;
			}
			else
			{
				encoding.modRM.mod = Encoding::MOD_DWORD_DISP;
				encoding.format.D1 = true;
				encoding.format.D2 = true;
				encoding.format.D3 = true;
				encoding.format.D4 = true;
			}
		}
		else
		{
			throw Error::INTERNAL;
		}
	}

	void Synthesizer::encodeSibByte()
	{
		if(scale == 0 && indexReg == Encoding::REG_UNKNOWN)
		{
			if((baseReg == Encoding::REG_UNKNOWN) ||
			   (encoding.modRM.r_m != Encoding::ESP && encoding.modRM.r_m != Encoding::EBP))
			{
				if(encoding.format.SIB)
				{
					throw Error::INTERNAL;
				}

				return;   // No SIB byte needed
			}
		}

		encoding.format.SIB = true;

		encoding.modRM.r_m = Encoding::ESP;   // Indicates use of SIB in mod R/M

		if(baseReg == Encoding::EBP && encoding.modRM.mod == Encoding::MOD_NO_DISP)
		{
			encoding.modRM.mod = Encoding::MOD_BYTE_DISP;

			encoding.format.D1 = true;
		}

		if(indexReg == Encoding::ESP)
		{
			if(scale != 1)
			{
				throw Error("ESP can't be scaled index in memory reference");
			}
			else   // Switch base and index
			{
				Encoding::Reg tempReg;

				tempReg = indexReg;
				indexReg = baseReg;
				baseReg = tempReg;
			}
		}

		if(baseReg == Encoding::REG_UNKNOWN)
		{
			encoding.SIB.base = Encoding::EBP;   // No Base

			encoding.modRM.mod = Encoding::MOD_NO_DISP;
			encoding.format.D1 = true;
			encoding.format.D2 = true;
			encoding.format.D3 = true;
			encoding.format.D4 = true;
		}
		else
		{
			encoding.SIB.base = baseReg;
		}

		if(indexReg != Encoding::REG_UNKNOWN)
		{
			encoding.SIB.index = indexReg;
		}
		else
		{
			encoding.SIB.index = Encoding::ESP;
		}

		switch(scale)
		{
		case 0:
		case 1:
			encoding.SIB.scale = Encoding::SCALE_1;
			break;
		case 2:
			encoding.SIB.scale = Encoding::SCALE_2;
			break;
		case 4:
			encoding.SIB.scale = Encoding::SCALE_4;
			break;
		case 8:
			encoding.SIB.scale = Encoding::SCALE_8;
			break;
		default:
			throw Error::INTERNAL;
		}
	}
}